"""Pyscopg helpers to work with PostGIS."""

VERSION = (0, 0, 2)

__author__ = 'Yohan Boniface, Zhou Qi'
__contact__ = "yohan.boniface@data.gouv.fr, bitwolaiye@gmail.com"
__homepage__ = "https://github.com/bitwolaiye/psycopg-postgis"
__version__ = ".".join(map(str, VERSION))



from psycopg2 import extensions as _ext
from .reader import EWKBReader, Typed
from .geojson import GeoJSON


class Geometry(object):
    __metaclass__ = Typed
    @property
    def has_z(self):
        return False

    @property
    def has_m(self):
        return False

    @property
    def has_srid(self):
        return hasattr(self, 'srid')

    @staticmethod
    def from_ewkb(value, cursor=None):
        if not value:
            return None
        return EWKBReader.from_hex(value)

    def __conform__(self, protocol):
        if protocol is _ext.ISQLQuote:
            return self

    def getquoted(self):
        return "ST_GeometryFromText('{}', {})".format(self.wkt, self.srid)

    def __str__(self):
        return self.wkt

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.wkt)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            other = other.coords
        return self.coords == other

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def wkt(self):
        return "{}({})".format(self.name.upper(), self.wkt_coords)

    @property
    def geojson(self):
        return GeoJSON({
            'type': self.name,
            'coordinates': self.coords
        })


def register(cursor):
    cursor.execute("SELECT NULL::geometry")
    oid = cursor.description[0][1]
    GEOMETRY = _ext.new_type((oid, ), "GEOMETRY", Geometry.from_ewkb)
    _ext.register_type(GEOMETRY)



from .geometry import Geometry
from .linestring import LineString


class Polygon(Geometry):

    TYPE = 3

    def __init__(self, rings, srid=None):
        self.rings = rings
        if srid:
            self.srid = srid

    def __iter__(self):
        for line in self.rings:
            if not isinstance(line, LineString):
                line = LineString(line)
            yield line

    @property
    def has_z(self):
        return self.rings[0].has_z

    @property
    def has_m(self):
        return self.rings[0].has_m

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls([LineString.from_ewkb_body(reader)
                   for index in range(reader.read_int())], srid)

    @property
    def wkt_coords(self):
        return ', '.join('({})'.format(r.wkt_coords) for r in self)

    @property
    def coords(self):
        return tuple(r.coords for r in self)



"Postgis helpers for psycopg2."
from .geometry import Geometry, register
from .point import Point
from .linestring import LineString
from .polygon import Polygon
from .multipoint import MultiPoint
from .multilinestring import MultiLineString
from .multipolygon import MultiPolygon
from .geometrycollection import GeometryCollection
from .__meta__ import __version__

__all__ = ['Geometry', 'register', 'Point', 'LineString', 'Polygon',
           'MultiPoint', 'MultiLineString', 'MultiPolygon',
           'GeometryCollection', '__version__']



from .geometry import Geometry
from .point import Point


class LineString(Geometry):

    TYPE = 2

    def __init__(self, points, srid=None):

        self.points = [self._ensure_point(p) for p in points]

        if srid:
            self.srid = srid

    def _ensure_point(self, point):
        if not isinstance(point, Point):
            return Point(*point)
        else:
            return point

    def __iter__(self):
        for point in self.points:
            # if not isinstance(point, Point):
            #     point = Point(*point)
            yield point

    @property
    def has_z(self):
        return self.points[0].has_z

    @property
    def has_m(self):
        return self.points[0].has_m

    def __getitem__(self, item):
        return self.points[item]

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls([Point.from_ewkb_body(reader)
                   for index in range(reader.read_int())], srid)

    @property
    def wkt_coords(self):
        return ', '.join(p.wkt_coords for p in self)

    @property
    def coords(self):
        return tuple(p.coords for p in self)



from .geometry import Geometry


class Point(Geometry):

    TYPE = 1

    def __init__(self, x, y=None, z=None, m=None, srid=None):
        if y is None and isinstance(x, (tuple, list)):
            if len(x) == 2:
                x, y = x
            elif len(x) == 3:
                x, y, z = x
            elif len(x) == 4:
                x, y, z, m = x
            # x, y, *extra = x
            # if extra:
            #     z, *extra = extra
            #     if extra:
            #         m = extra[0]
        self.x = x
        self.y = y
        self.z = z
        self.m = m
        if self.x is None:
            self.key_len = 0
        elif self.y is None:
            self.key_len = 1
        elif self.z is None:
            self.key_len = 2
        elif self.m is None:
            self.key_len = 3
        else:
            self.key_len = 4
        if srid is not None:
            self.srid = srid

    def __getitem__(self, item):
        if item in (0, 'x'):
            return self.x
        elif item in (1, 'y'):
            return self.y
        elif item in (2, 'z'):
            return self.z
        elif item in (3, 'm'):
            return self.m

    def __iter__(self):
        return iter(self.values())

    _keys = ['x', 'y', 'z', 'm']

    def keys(self):
        return [k for k in self._keys if self[k] is not None]

    def values(self):
        return tuple(self[k] for k in xrange(self.key_len))

    @property
    def has_z(self):
        return self.z is not None

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls(reader.read_double(), reader.read_double(),
                   reader.read_double() if reader.has_z else None,
                   reader.read_double() if reader.has_m else None,
                   srid)

    @property
    def wkt_coords(self):
        return ' '.join(map(str, self.coords))

    @property
    def coords(self):
        if self.key_len == 2:
            return self.x, self.y
        elif self.key_len == 3:
            return self.x, self.y, self.z
        elif self.key_len == 4:
            return self.x, self.y, self.z, self.m
        elif self.key_len == 1:
            return self.x,
        else:
            return ()



import binascii
from io import BytesIO
import struct


class Typed(type):

    types = {}

    def __new__(mcs, name, bases, attrs, **kwargs):
        cls = super(Typed, mcs).__new__(mcs, name, bases, attrs, **kwargs)
        if hasattr(cls, 'TYPE'):
            Typed.types[cls.TYPE] = cls
        return cls


class EWKBReader(object):
    # Should we use shapely/libgeos instead?

    def __init__(self, stream):
        self.stream = stream

    def clone(self):
        return EWKBReader(self.stream)

    def read(self):
        # https://en.wikipedia.org/wiki/Well-known_text#Well-known_binary
        byte_order = self.stream.read(1)
        if byte_order == b'\x00':
            self._endianness = b'>'
        elif byte_order == b'\x01':
            self._endianness = b'<'
        else:
            raise Exception('invalid EWKB encoding')

        type_ = self.read_int()
        self.has_z = bool(type_ & 0x80000000)
        self.has_m = bool(type_ & 0x40000000)
        srid = self.read_int() if bool(type_ & 0x20000000) else None
        type_ &= 0x1fffffff

        try:
            class_ = Typed.types[type_]
        except KeyError:
            raise ValueError('unsupported geometry type {0}'.format(type_))
        else:
            return class_.from_ewkb_body(self, srid)

    def read_int(self):
        return struct.unpack(self._endianness + b'I', self.stream.read(4))[0]

    def read_double(self):
        return struct.unpack(self._endianness + b'd', self.stream.read(8))[0]

    @classmethod
    def from_hex(cls, value):
        return cls(BytesIO(binascii.a2b_hex(value))).read()
